# THIS FILE IS GENERATED FROM macrosynergy SETUP.PY
short_version: str = '0.0.38'
version: str = '0.0.38'
full_version: str = '0.0.38'
git_revision: str = '207ea27d3fcf8c0721da3958821581c408d5ba9f'
release: bool = True
if not release:
    version = full_version
